from __future__ import annotations
from librespot import util
from librespot.audio.decrypt import AesAudioDecrypt
from librespot.audio.format import SuperAudioFormat
from librespot.audio.storage import ChannelManager
from librespot.cache import CacheManager
from librespot.crypto import Packet
from librespot.metadata import EpisodeId, PlayableId, TrackId
from librespot.proto import Metadata_pb2 as Metadata, StorageResolve_pb2 as StorageResolve
from librespot.structure import AudioDecrypt, AudioQualityPicker, Closeable, FeederException, GeneralAudioStream, GeneralWritableStream, HaltListener, NoopAudioDecrypt, PacketsReceiver
import concurrent.futures
import io
import logging
import math
import queue
import random
import struct
import threading
import time
import typing
import urllib.parse

if typing.TYPE_CHECKING:
    from librespot.core import Session


class AbsChunkedInputStream(io.BytesIO, HaltListener):
    chunk_exception = None
    closed = False
    max_chunk_tries = 128
    preload_ahead = 3
    preload_chunk_retries = 2
    retries: typing.List[int]
    retry_on_chunk_error: bool
    wait_lock: threading.Condition = threading.Condition()
    wait_for_chunk = -1
    __decoded_length = 0
    __mark = 0
    __pos = 0

    def __init__(self, retry_on_chunk_error: bool):
        super().__init__()
        self.retries = [0] * self.chunks()
        self.retry_on_chunk_error = retry_on_chunk_error

    def is_closed(self) -> bool:
        return self.closed

    def buffer(self) -> typing.List[bytes]:
        raise NotImplementedError()

    def size(self) -> int:
        raise NotImplementedError()

    def close(self) -> None:
        self.closed = True
        with self.wait_lock:
            self.wait_lock.notify_all()

    def available(self):
        return self.size() - self.__pos

    def mark_supported(self) -> bool:
        return True

    def mark(self, read_ahead_limit: int) -> None:
        self.__mark = self.__pos

    def reset(self) -> None:
        self.__pos = self.__mark

    def pos(self) -> int:
        return self.__pos

    def seek(self, where: int, **kwargs) -> None:
        if where < 0:
            raise TypeError()
        if self.closed:
            raise IOError("Stream is closed!")
        self.__pos = where
        self.check_availability(int(self.__pos / (128 * 1024)), False, False)

    def skip(self, n: int) -> int:
        if n < 0:
            raise TypeError()
        if self.closed:
            raise IOError("Stream is closed!")
        k = self.size() - self.__pos
        if n < k:
            k = n
        self.__pos += k
        chunk = int(self.__pos / (128 * 1024))
        self.check_availability(chunk, False, False)
        return k

    def requested_chunks(self) -> typing.List[bool]:
        raise NotImplementedError()

    def available_chunks(self) -> typing.List[bool]:
        raise NotImplementedError()

    def chunks(self) -> int:
        raise NotImplementedError()

    def request_chunk_from_stream(self, index: int) -> None:
        raise NotImplementedError()

    def should_retry(self, chunk: int) -> bool:
        if self.retries[chunk] < 1:
            return True
        if self.retries[chunk] > self.max_chunk_tries:
            return False
        return self.retry_on_chunk_error

    def check_availability(self, chunk: int, wait: bool, halted: bool) -> None:
        if halted and not wait:
            raise TypeError()
        if not self.requested_chunks()[chunk]:
            self.request_chunk_from_stream(chunk)
            self.requested_chunks()[chunk] = True
        for i in range(chunk + 1,
                       min(self.chunks() - 1, chunk + self.preload_ahead) + 1):
            if (self.requested_chunks()[i]
                    and self.retries[i] < self.preload_chunk_retries):
                self.request_chunk_from_stream(i)
                self.requested_chunks()[chunk] = True
        if wait:
            if self.available_chunks()[chunk]:
                return
            retry = False
            with self.wait_lock:
                if not halted:
                    self.stream_read_halted(chunk, int(time.time() * 1000))
                self.chunk_exception = None
                self.wait_for_chunk = chunk
                self.wait_lock.wait_for(lambda: self.available_chunks()[chunk])
                if self.closed:
                    return
                if self.chunk_exception is not None:
                    if self.should_retry(chunk):
                        retry = True
                    else:
                        raise AbsChunkedInputStream.ChunkException
                if not retry:
                    self.stream_read_halted(chunk, int(time.time() * 1000))
            if retry:
                time.sleep(math.log10(self.retries[chunk]))
                self.check_availability(chunk, True, True)

    def read(self, __size: int = 0) -> bytes:
        if self.closed:
            raise IOError("Stream is closed!")
        if __size <= 0:
            if self.__pos == self.size():
                return b""
            buffer = io.BytesIO()
            total_size = self.size()
            chunk = int(self.__pos / (128 * 1024))
            chunk_off = int(self.__pos % (128 * 1024))
            chunk_total = int(math.ceil(total_size / (128 * 1024)))
            self.check_availability(chunk, True, False)
            buffer.write(self.buffer()[chunk][chunk_off:])
            chunk += 1
            if chunk != chunk_total:
                while chunk <= chunk_total - 1:
                    self.check_availability(chunk, True, False)
                    buffer.write(self.buffer()[chunk])
                    chunk += 1
            buffer.seek(0)
            self.__pos += buffer.getbuffer().nbytes
            return buffer.read()
        buffer = io.BytesIO()
        chunk = int(self.__pos / (128 * 1024))
        chunk_off = int(self.__pos % (128 * 1024))
        chunk_end = int(__size / (128 * 1024))
        chunk_end_off = int(__size % (128 * 1024))
        if chunk_end > self.size():
            chunk_end = int(self.size() / (128 * 1024))
            chunk_end_off = int(self.size() % (128 * 1024))
        self.check_availability(chunk, True, False)
        if chunk_off + __size > len(self.buffer()[chunk]):
            buffer.write(self.buffer()[chunk][chunk_off:])
            chunk += 1
            while chunk <= chunk_end:
                self.check_availability(chunk, True, False)
                if chunk == chunk_end:
                    buffer.write(self.buffer()[chunk][:chunk_end_off])
                else:
                    buffer.write(self.buffer()[chunk])
                chunk += 1
        else:
            buffer.write(self.buffer()[chunk][chunk_off:chunk_off + __size])
        buffer.seek(0)
        self.__pos += buffer.getbuffer().nbytes
        return buffer.read()

    def notify_chunk_available(self, index: int) -> None:
        self.available_chunks()[index] = True
        self.__decoded_length += len(self.buffer()[index])
        with self.wait_lock:
            if index == self.wait_for_chunk and not self.closed:
                self.wait_for_chunk = -1
                self.wait_lock.notify_all()

    def notify_chunk_error(self, index: int, ex):
        self.available_chunks()[index] = False
        self.requested_chunks()[index] = False
        self.retries[index] += 1
        with self.wait_lock:
            if index == self.wait_for_chunk and not self.closed:
                self.chunk_exception = ex
                self.wait_for_chunk = -1
                self.wait_lock.notify_all()

    def decoded_length(self):
        return self.__decoded_length

    class ChunkException(IOError):

        @staticmethod
        def from_stream_error(stream_error: int):
            return AbsChunkedInputStream \
                .ChunkException("Failed due to stream error, code: {}".format(stream_error))


class AudioKeyManager(PacketsReceiver, Closeable):
    audio_key_request_timeout = 20
    logger = logging.getLogger("Librespot:AudioKeyManager")
    __callbacks: typing.Dict[int, Callback] = {}
    __seq_holder = 0
    __seq_holder_lock = threading.Condition()
    __session: Session
    __zero_short = b"\x00\x00"

    def __init__(self, session: Session):
        self.__session = session

    def dispatch(self, packet: Packet) -> None:
        payload = io.BytesIO(packet.payload)
        seq = struct.unpack(">i", payload.read(4))[0]
        callback = self.__callbacks.get(seq)
        if callback is None:
            self.logger.warning(
                "Couldn't find callback for seq: {}".format(seq))
            return
        if packet.is_cmd(Packet.Type.aes_key):
            key = payload.read(16)
            callback.key(key)
        elif packet.is_cmd(Packet.Type.aes_key_error):
            code = struct.unpack(">H", payload.read(2))[0]
            callback.error(code)
        else:
            self.logger.warning(
                "Couldn't handle packet, cmd: {}, length: {}".format(
                    packet.cmd, len(packet.payload)))

    def get_audio_key(self,
                      gid: bytes,
                      file_id: bytes,
                      retry: bool = True) -> bytes:
        seq: int
        with self.__seq_holder_lock:
            seq = self.__seq_holder
            self.__seq_holder += 1
        out = io.BytesIO()
        out.write(file_id)
        out.write(gid)
        out.write(struct.pack(">i", seq))
        out.write(self.__zero_short)
        out.seek(0)
        self.__session.send(Packet.Type.request_key, out.read())
        callback = AudioKeyManager.SyncCallback(self)
        self.__callbacks[seq] = callback
        key = callback.wait_response()
        if key is None:
            if retry:
                return self.get_audio_key(gid, file_id, False)
            raise RuntimeError(
                "Failed fetching audio key! gid: {}, fileId: {}".format(
                    util.bytes_to_hex(gid), util.bytes_to_hex(file_id)))
        return key

    class Callback:

        def key(self, key: bytes) -> None:
            raise NotImplementedError

        def error(self, code: int) -> None:
            raise NotImplementedError

    class SyncCallback(Callback):
        __audio_key_manager: AudioKeyManager
        __reference = queue.Queue()
        __reference_lock = threading.Condition()

        def __init__(self, audio_key_manager: AudioKeyManager):
            self.__audio_key_manager = audio_key_manager

        def key(self, key: bytes) -> None:
            with self.__reference_lock:
                self.__reference.put(key)
                self.__reference_lock.notify_all()

        def error(self, code: int) -> None:
            self.__audio_key_manager.logger.fatal(
                "Audio key error, code: {}".format(code))
            with self.__reference_lock:
                self.__reference.put(None)
                self.__reference_lock.notify_all()

        def wait_response(self) -> bytes:
            with self.__reference_lock:
                self.__reference_lock.wait(
                    AudioKeyManager.audio_key_request_timeout)
                return self.__reference.get(block=False)


class CdnFeedHelper:
    _LOGGER: logging = logging.getLogger(__name__)

    @staticmethod
    def get_url(resp: StorageResolve.StorageResolveResponse) -> str:
        selected_url = random.choice(resp.cdnurl)
        while "audio4-gm-fb" in selected_url or "audio-gm-fb" in selected_url:
            selected_url = random.choice(resp.cdnurl)
        return selected_url

    @staticmethod
    def load_track(
            session: Session, track: Metadata.Track, file: Metadata.AudioFile,
            resp_or_url: typing.Union[StorageResolve.StorageResolveResponse,
                                      str], preload: bool,
            halt_listener: HaltListener) -> PlayableContentFeeder.LoadedStream:
        if type(resp_or_url) is str:
            url = resp_or_url
        else:
            url = CdnFeedHelper.get_url(resp_or_url)
        start = int(time.time() * 1000)
        key = session.audio_key().get_audio_key(track.gid, file.file_id)
        audio_key_time = int(time.time() * 1000) - start

        streamer = session.cdn().stream_file(file, key, url, halt_listener)
        input_stream = streamer.stream()
        normalization_data = NormalizationData.read(input_stream)
        if input_stream.skip(0xA7) != 0xA7:
            raise IOError("Couldn't skip 0xa7 bytes!")
        return PlayableContentFeeder.LoadedStream(
            track,
            streamer,
            normalization_data,
            PlayableContentFeeder.Metrics(file.file_id, preload,
                                          -1 if preload else audio_key_time),
        )

    @staticmethod
    def load_episode_external(
            session: Session, episode: Metadata.Episode,
            halt_listener: HaltListener) -> PlayableContentFeeder.LoadedStream:
        resp = session.client().head(episode.external_url)

        if resp.status_code != 200:
            CdnFeedHelper._LOGGER.warning("Couldn't resolve redirect!")

        url = resp.url
        CdnFeedHelper._LOGGER.debug("Fetched external url for {}: {}".format(
            util.bytes_to_hex(episode.gid), url))

        streamer = session.cdn().stream_external_episode(
            episode, url, halt_listener)
        return PlayableContentFeeder.LoadedStream(
            episode,
            streamer,
            None,
            PlayableContentFeeder.Metrics(None, False, -1),
        )

    @staticmethod
    def load_episode(
        session: Session,
        episode: Metadata.Episode,
        file: Metadata.AudioFile,
        resp_or_url: typing.Union[StorageResolve.StorageResolveResponse, str],
        preload: bool,
        halt_listener: HaltListener,
    ) -> PlayableContentFeeder.LoadedStream:
        if type(resp_or_url) is str:
            url = resp_or_url
        else:
            url = CdnFeedHelper.get_url(resp_or_url)
        start = int(time.time() * 1000)
        key = session.audio_key().get_audio_key(episode.gid, file.file_id)
        audio_key_time = int(time.time() * 1000) - start

        streamer = session.cdn().stream_file(file, key, url, halt_listener)
        input_stream = streamer.stream()
        normalization_data = NormalizationData.read(input_stream)
        if input_stream.skip(0xA7) != 0xA7:
            raise IOError("Couldn't skip 0xa7 bytes!")
        return PlayableContentFeeder.LoadedStream(
            episode,
            streamer,
            normalization_data,
            PlayableContentFeeder.Metrics(file.file_id, preload,
                                          -1 if preload else audio_key_time),
        )


class CdnManager:
    logger: logging = logging.getLogger("Librespot:CdnManager")
    __session: Session

    def __init__(self, session: Session):
        self.__session = session

    def get_head(self, file_id: bytes):
        response = self.__session.client() \
            .get(self.__session.get_user_attribute("head-files-url", "https://heads-fa.spotify.com/head/{file_id}")
                 .replace("{file_id}", util.bytes_to_hex(file_id)))
        if response.status_code != 200:
            raise IOError("{}".format(response.status_code))
        body = response.content
        if body is None:
            raise IOError("Response body is empty!")
        return body

    def stream_external_episode(self, episode: Metadata.Episode,
                                external_url: str,
                                halt_listener: HaltListener):
        return CdnManager.Streamer(
            self.__session,
            StreamId(episode=episode),
            SuperAudioFormat.MP3,
            CdnManager.CdnUrl(self, None, external_url),
            self.__session.cache(),
            NoopAudioDecrypt(),
            halt_listener,
        )

    def stream_file(self, file: Metadata.AudioFile, key: bytes, url: str,
                    halt_listener: HaltListener):
        return CdnManager.Streamer(
            self.__session,
            StreamId(file=file),
            SuperAudioFormat.get(file.format),
            CdnManager.CdnUrl(self, file.file_id, url),
            self.__session.cache(),
            AesAudioDecrypt(key),
            halt_listener,
        )

    def get_audio_url(self, file_id: bytes):
        response = self.__session.api()\
            .send("GET", "/storage-resolve/files/audio/interactive/{}".format(util.bytes_to_hex(file_id)), None, None)
        if response.status_code != 200:
            raise IOError(response.status_code)
        body = response.content
        if body is None:
            raise IOError("Response body is empty!")
        proto = StorageResolve.StorageResolveResponse()
        proto.ParseFromString(body)
        if proto.result == StorageResolve.StorageResolveResponse.Result.CDN:
            url = random.choice(proto.cdnurl)
            self.logger.debug("Fetched CDN url for {}: {}".format(
                util.bytes_to_hex(file_id), url))
            return url
        raise CdnManager.CdnException(
            "Could not retrieve CDN url! result: {}".format(proto.result))

    class CdnException(Exception):
        pass

    class InternalResponse:
        buffer: bytes
        headers: typing.Dict[str, str]

        def __init__(self, buffer: bytes, headers: typing.Dict[str, str]):
            self.buffer = buffer
            self.headers = headers

    class CdnUrl:
        __cdn_manager = None
        __file_id: bytes
        __expiration: int
        url: str

        def __init__(self, cdn_manager, file_id: typing.Union[bytes, None],
                     url: str):
            self.__cdn_manager: CdnManager = cdn_manager
            self.__file_id = file_id
            self.set_url(url)

        def url(self):
            if self.__expiration == -1:
                return self.url
            if self.__expiration <= int(time.time() * 1000) + 5 * 60 * 1000:
                self.url = self.__cdn_manager.get_audio_url(self.__file_id)
            return self.url

        def set_url(self, url: str):
            self.url = url
            if self.__file_id is not None:
                token_url = urllib.parse.urlparse(url)
                token_query = urllib.parse.parse_qs(token_url.query)
                token_list = token_query.get("__token__")
                try:
                    token_str = str(token_list[0])
                except TypeError:
                    token_str = ""
                expires_list = token_query.get("Expires")
                try:
                    expires_str = str(expires_list[0])
                except TypeError:
                    expires_str = ""
                if token_str != "None" and len(token_str) != 0:
                    expire_at = None
                    split = token_str.split("~")
                    for s in split:
                        try:
                            i = s.index("=")
                        except ValueError:
                            continue
                        if s[:i] == "exp":
                            expire_at = int(s[i + 1:])
                            break
                    if expire_at is None:
                        self.__expiration = -1
                        self.__cdn_manager.logger.warning(
                            "Invalid __token__ in CDN url: {}".format(url))
                        return
                    self.__expiration = expire_at * 1000
                elif expires_str != "None" and len(expires_str) != 0:
                    expires_at = None
                    expires_str = expires_str.split("~")[0]
                    expires_at = int(expires_str)
                    if expires_at is None:
                        self.__expiration = -1
                        self.__cdn_manager.logger.warning("Invalid Expires param in CDN url: {}".format(url))
                        return
                    self.__expiration = expires_at * 1000
                else:
                    try:
                        i = token_url.query.index("_")
                    except ValueError:
                        self.__expiration = -1
                        self.__cdn_manager.logger \
                            .warning("Couldn't extract expiration, invalid parameter in CDN url: {}".format(url))
                        return
                    self.__expiration = int(token_url.query[:i]) * 1000

            else:
                self.__expiration = -1

    class Streamer(GeneralAudioStream, GeneralWritableStream):
        available: typing.List[bool]
        buffer: typing.List[bytes]
        chunks: int
        executor_service = concurrent.futures.ThreadPoolExecutor()
        halt_listener: HaltListener
        requested: typing.List[bool]
        size: int
        __audio_format: SuperAudioFormat
        __audio_decrypt: AudioDecrypt
        __cdn_url: CdnManager.CdnUrl
        __internal_stream: InternalStream
        __session: Session
        __stream_id: StreamId

        def __init__(self, session: Session, stream_id: StreamId,
                     audio_format: SuperAudioFormat,
                     cdn_url: CdnManager.CdnUrl, cache: CacheManager,
                     audio_decrypt: AudioDecrypt, halt_listener: HaltListener):
            self.__session = session
            self.__stream_id = stream_id
            self.__audio_format = audio_format
            self.__audio_decrypt = audio_decrypt
            self.__cdn_url = cdn_url
            self.halt_listener = halt_listener
            response = self.request(range_start=0,
                                    range_end=ChannelManager.chunk_size - 1)
            content_range = response.headers.get("Content-Range")
            if content_range is None:
                raise IOError("Missing Content-Range header!")
            split = content_range.split("/")
            self.size = int(split[1])
            self.chunks = int(math.ceil(self.size / ChannelManager.chunk_size))
            first_chunk = response.buffer
            self.available = [False for _ in range(self.chunks)]
            self.requested = [False for _ in range(self.chunks)]
            self.buffer = [b"" for _ in range(self.chunks)]
            self.__internal_stream = CdnManager.Streamer.InternalStream(
                self, False)
            self.requested[0] = True
            self.write_chunk(first_chunk, 0, False)

        def write_chunk(self, chunk: bytes, chunk_index: int,
                        cached: bool) -> None:
            if self.__internal_stream.is_closed():
                return
            self.__session.logger.debug(
                "Chunk {}/{} completed, cached: {}, stream: {}".format(
                    chunk_index + 1, self.chunks, cached, self.describe()))
            self.buffer[chunk_index] = self.__audio_decrypt.decrypt_chunk(
                chunk_index, chunk)
            self.__internal_stream.notify_chunk_available(chunk_index)

        def stream(self) -> AbsChunkedInputStream:
            return self.__internal_stream

        def codec(self) -> SuperAudioFormat:
            return self.__audio_format

        def describe(self) -> str:
            if self.__stream_id.is_episode():
                return "episode_gid: {}".format(
                    self.__stream_id.get_episode_gid())
            return "file_id: {}".format(self.__stream_id.get_file_id())

        def decrypt_time_ms(self) -> int:
            return self.__audio_decrypt.decrypt_time_ms()

        def request_chunk(self, index: int) -> None:
            response = self.request(index)
            self.write_chunk(response.buffer, index, False)

        def request(self, chunk: int = None, range_start: int = None, range_end: int = None)\
                -> CdnManager.InternalResponse:
            if chunk is None and range_start is None and range_end is None:
                raise TypeError()
            if chunk is not None:
                range_start = ChannelManager.chunk_size * chunk
                range_end = (chunk + 1) * ChannelManager.chunk_size - 1
            response = self.__session.client().get(
                self.__cdn_url.url,
                headers={
                    "Range": "bytes={}-{}".format(range_start, range_end)
                },
            )
            if response.status_code != 206:
                raise IOError(response.status_code)
            body = response.content
            if body is None:
                raise IOError("Response body is empty!")
            return CdnManager.InternalResponse(body, dict(response.headers))

        class InternalStream(AbsChunkedInputStream):
            streamer: CdnManager.Streamer

            def __init__(self, streamer, retry_on_chunk_error: bool):
                self.streamer: CdnManager.Streamer = streamer
                super().__init__(retry_on_chunk_error)

            def buffer(self) -> typing.List[bytes]:
                return self.streamer.buffer

            def size(self) -> int:
                return self.streamer.size

            def close(self) -> None:
                super().close()
                del self.streamer.buffer

            def requested_chunks(self) -> typing.List[bool]:
                return self.streamer.requested

            def available_chunks(self) -> typing.List[bool]:
                return self.streamer.available

            def chunks(self) -> int:
                return self.streamer.chunks

            def request_chunk_from_stream(self, index: int) -> None:
                self.streamer.executor_service \
                    .submit(lambda: self.streamer.request_chunk(index))

            def stream_read_halted(self, chunk: int, _time: int) -> None:
                if self.streamer.halt_listener is not None:
                    self.streamer.executor_service\
                        .submit(lambda: self.streamer.halt_listener.stream_read_halted(chunk, _time))

            def stream_read_resumed(self, chunk: int, _time: int) -> None:
                if self.streamer.halt_listener is not None:
                    self.streamer.executor_service \
                        .submit(lambda: self.streamer.halt_listener.stream_read_resumed(chunk, _time))


class NormalizationData:
    _LOGGER: logging = logging.getLogger(__name__)
    track_gain_db: float
    track_peak: float
    album_gain_db: float
    album_peak: float

    def __init__(self, track_gain_db: float, track_peak: float,
                 album_gain_db: float, album_peak: float):
        self.track_gain_db = track_gain_db
        self.track_peak = track_peak
        self.album_gain_db = album_gain_db
        self.album_peak = album_peak

        self._LOGGER.debug(
            "Loaded normalization data, track_gain: {}, track_peak: {}, album_gain: {}, album_peak: {}"
            .format(track_gain_db, track_peak, album_gain_db, album_peak))

    @staticmethod
    def read(input_stream: AbsChunkedInputStream) -> NormalizationData:
        input_stream.seek(144)
        data = input_stream.read(4 * 4)
        input_stream.seek(0)
        buffer = io.BytesIO(data)
        return NormalizationData(
            struct.unpack("<f", buffer.read(4))[0],
            struct.unpack("<f", buffer.read(4))[0],
            struct.unpack("<f", buffer.read(4))[0],
            struct.unpack("<f", buffer.read(4))[0])

    def get_factor(self, normalisation_pregain) -> float:
        normalisation_factor = float(
            math.pow(10, (self.track_gain_db + normalisation_pregain) / 20))
        if normalisation_factor * self.track_peak > 1:
            self._LOGGER \
                .warning("Reducing normalisation factor to prevent clipping. Please add negative pregain to avoid.")
            normalisation_factor = 1 / self.track_peak
        return normalisation_factor


class PlayableContentFeeder:
    logger = logging.getLogger("Librespot:PlayableContentFeeder")
    storage_resolve_interactive = "/storage-resolve/files/audio/interactive/{}"
    storage_resolve_interactive_prefetch = "/storage-resolve/files/audio/interactive_prefetch/{}"
    __session: Session

    def __init__(self, session: Session):
        self.__session = session

    def load(self, playable_id: PlayableId,
             audio_quality_picker: AudioQualityPicker, preload: bool,
             halt_listener: typing.Union[HaltListener, None]):
        if type(playable_id) is TrackId:
            return self.load_track(playable_id, audio_quality_picker, preload,
                                   halt_listener)
        if type(playable_id) is EpisodeId:
            return self.load_episode(playable_id, audio_quality_picker,
                                     preload, halt_listener)
        raise TypeError("Unknown content: {}".format(playable_id))

    def load_stream(self, file: Metadata.AudioFile, track: Metadata.Track,
                    episode: Metadata.Episode, preload: bool,
                    halt_lister: HaltListener):
        if track is None and episode is None:
            raise RuntimeError()
        response = self.resolve_storage_interactive(file.file_id, preload)
        if response.result == StorageResolve.StorageResolveResponse.Result.CDN:
            if track is not None:
                return CdnFeedHelper.load_track(self.__session, track, file,
                                                response, preload, halt_lister)
            return CdnFeedHelper.load_episode(self.__session, episode, file,
                                              response, preload, halt_lister)
        if response.result == StorageResolve.StorageResolveResponse.Result.STORAGE:
            if track is None:
                pass
        elif response.result == StorageResolve.StorageResolveResponse.Result.RESTRICTED:
            raise RuntimeError("Content is restricted!")
        elif response.result == StorageResolve.StorageResolveResponse.Response.UNRECOGNIZED:
            raise RuntimeError("Content is unrecognized!")
        else:
            raise RuntimeError("Unknown result: {}".format(response.result))

    def load_episode(self, episode_id: EpisodeId,
                     audio_quality_picker: AudioQualityPicker, preload: bool,
                     halt_listener: HaltListener) -> LoadedStream:
        episode = self.__session.api().get_metadata_4_episode(episode_id)
        if episode.external_url:
            return CdnFeedHelper.load_episode_external(self.__session, episode,
                                                       halt_listener)
        file = audio_quality_picker.get_file(episode.audio)
        if file is None:
            self.logger.fatal(
                "Couldn't find any suitable audio file, available: {}".format(
                    episode.audio))
        return self.load_stream(file, None, episode, preload, halt_listener)

    def load_track(self, track_id_or_track: typing.Union[TrackId,
                                                         Metadata.Track],
                   audio_quality_picker: AudioQualityPicker, preload: bool,
                   halt_listener: HaltListener):
        if type(track_id_or_track) is TrackId:
            original = self.__session.api().get_metadata_4_track(
                track_id_or_track)
            track = self.pick_alternative_if_necessary(original)
            if track is None:
                raise RuntimeError("Cannot get alternative track")
        else:
            track = track_id_or_track
        file = audio_quality_picker.get_file(track.file)
        if file is None:
            self.logger.fatal(
                "Couldn't find any suitable audio file, available: {}".format(
                    track.file))
            raise FeederException()
        return self.load_stream(file, track, None, preload, halt_listener)

    def pick_alternative_if_necessary(
            self, track: Metadata.Track) -> typing.Union[Metadata.Track, None]:
        if len(track.file) > 0:
            return track
        for alt in track.alternative:
            if len(alt.file) > 0:
                return Metadata.Track(
                    gid=track.gid,
                    name=track.name,
                    album=track.album,
                    artist=track.artist,
                    number=track.number,
                    disc_number=track.disc_number,
                    duration=track.duration,
                    popularity=track.popularity,
                    explicit=track.explicit,
                    external_id=track.external_id,
                    restriction=track.restriction,
                    file=alt.file,
                    sale_period=track.sale_period,
                    preview=track.preview,
                    tags=track.tags,
                    earliest_live_timestamp=track.earliest_live_timestamp,
                    has_lyrics=track.has_lyrics,
                    availability=track.availability,
                    licensor=track.licensor)
        return None

    def resolve_storage_interactive(
            self, file_id: bytes,
            preload: bool) -> StorageResolve.StorageResolveResponse:
        resp = self.__session.api().send(
            "GET",
            (self.storage_resolve_interactive_prefetch
             if preload else self.storage_resolve_interactive).format(
                 util.bytes_to_hex(file_id)),
            None,
            None,
        )
        if resp.status_code != 200:
            raise RuntimeError(resp.status_code)
        body = resp.content
        if body is None:
            raise RuntimeError("Response body is empty!")
        storage_resolve_response = StorageResolve.StorageResolveResponse()
        storage_resolve_response.ParseFromString(body)
        return storage_resolve_response

    class LoadedStream:
        episode: Metadata.Episode
        track: Metadata.Track
        input_stream: GeneralAudioStream
        normalization_data: NormalizationData
        metrics: PlayableContentFeeder.Metrics

        def __init__(self, track_or_episode: typing.Union[Metadata.Track,
                                                          Metadata.Episode],
                     input_stream: GeneralAudioStream,
                     normalization_data: typing.Union[NormalizationData, None],
                     metrics: PlayableContentFeeder.Metrics):
            if type(track_or_episode) is Metadata.Track:
                self.track = track_or_episode
                self.episode = None
            elif type(track_or_episode) is Metadata.Episode:
                self.track = None
                self.episode = track_or_episode
            else:
                raise TypeError()
            self.input_stream = input_stream
            self.normalization_data = normalization_data
            self.metrics = metrics

    class Metrics:
        file_id: str
        preloaded_audio_key: bool
        audio_key_time: int

        def __init__(self, file_id: typing.Union[bytes, None],
                     preloaded_audio_key: bool, audio_key_time: int):
            self.file_id = None if file_id is None else util.bytes_to_hex(
                file_id)
            self.preloaded_audio_key = preloaded_audio_key
            self.audio_key_time = audio_key_time
            if preloaded_audio_key and audio_key_time != -1:
                raise RuntimeError()


class StreamId:
    file_id: bytes
    episode_gid: bytes

    def __init__(self,
                 file: Metadata.AudioFile = None,
                 episode: Metadata.Episode = None):
        if file is None and episode is None:
            return
        self.file_id = None if file is None else file.file_id
        self.episode_gid = None if episode is None else episode.gid

    def get_file_id(self):
        if self.file_id is None:
            raise RuntimeError("Not a file!")
        return util.bytes_to_hex(self.file_id)

    def is_episode(self):
        return self.episode_gid is not None

    def get_episode_gid(self):
        if self.episode_gid is None:
            raise RuntimeError("Not an episode!")
        return util.bytes_to_hex(self.episode_gid)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mercury.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='mercury.proto',
    package='spotify',
    syntax='proto2',
    serialized_options=b'\n\013com.spotify',
    create_key=_descriptor._internal_create_key,
    serialized_pb=
    b'\n\rmercury.proto\x12\x07spotify\"B\n\x16MercuryMultiGetRequest\x12(\n\x07request\x18\x01 \x03(\x0b\x32\x17.spotify.MercuryRequest\"<\n\x14MercuryMultiGetReply\x12$\n\x05reply\x18\x01 \x03(\x0b\x32\x15.spotify.MercuryReply\"O\n\x0eMercuryRequest\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x14\n\x0c\x63ontent_type\x18\x02 \x01(\t\x12\x0c\n\x04\x62ody\x18\x03 \x01(\x0c\x12\x0c\n\x04\x65tag\x18\x04 \x01(\x0c\"\xf5\x01\n\x0cMercuryReply\x12\x13\n\x0bstatus_code\x18\x01 \x01(\x11\x12\x16\n\x0estatus_message\x18\x02 \x01(\t\x12\x37\n\x0c\x63\x61\x63he_policy\x18\x03 \x01(\x0e\x32!.spotify.MercuryReply.CachePolicy\x12\x0b\n\x03ttl\x18\x04 \x01(\x11\x12\x0c\n\x04\x65tag\x18\x05 \x01(\x0c\x12\x14\n\x0c\x63ontent_type\x18\x06 \x01(\t\x12\x0c\n\x04\x62ody\x18\x07 \x01(\x0c\"@\n\x0b\x43\x61\x63hePolicy\x12\x0c\n\x08\x43\x41\x43HE_NO\x10\x01\x12\x11\n\rCACHE_PRIVATE\x10\x02\x12\x10\n\x0c\x43\x41\x43HE_PUBLIC\x10\x03\"y\n\x06Header\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x14\n\x0c\x63ontent_type\x18\x02 \x01(\t\x12\x0e\n\x06method\x18\x03 \x01(\t\x12\x13\n\x0bstatus_code\x18\x04 \x01(\x11\x12\'\n\x0buser_fields\x18\x06 \x03(\x0b\x32\x12.spotify.UserField\"\'\n\tUserField\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x42\r\n\x0b\x63om.spotify'
)

_MERCURYREPLY_CACHEPOLICY = _descriptor.EnumDescriptor(
    name='CachePolicy',
    full_name='spotify.MercuryReply.CachePolicy',
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name='CACHE_NO',
            index=0,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='CACHE_PRIVATE',
            index=1,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
        _descriptor.EnumValueDescriptor(
            name='CACHE_PUBLIC',
            index=2,
            number=3,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=419,
    serialized_end=483,
)
_sym_db.RegisterEnumDescriptor(_MERCURYREPLY_CACHEPOLICY)

_MERCURYMULTIGETREQUEST = _descriptor.Descriptor(
    name='MercuryMultiGetRequest',
    full_name='spotify.MercuryMultiGetRequest',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='request',
            full_name='spotify.MercuryMultiGetRequest.request',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=26,
    serialized_end=92,
)

_MERCURYMULTIGETREPLY = _descriptor.Descriptor(
    name='MercuryMultiGetReply',
    full_name='spotify.MercuryMultiGetReply',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='reply',
            full_name='spotify.MercuryMultiGetReply.reply',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=94,
    serialized_end=154,
)

_MERCURYREQUEST = _descriptor.Descriptor(
    name='MercuryRequest',
    full_name='spotify.MercuryRequest',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='uri',
            full_name='spotify.MercuryRequest.uri',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='content_type',
            full_name='spotify.MercuryRequest.content_type',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='body',
            full_name='spotify.MercuryRequest.body',
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='etag',
            full_name='spotify.MercuryRequest.etag',
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=156,
    serialized_end=235,
)

_MERCURYREPLY = _descriptor.Descriptor(
    name='MercuryReply',
    full_name='spotify.MercuryReply',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='status_code',
            full_name='spotify.MercuryReply.status_code',
            index=0,
            number=1,
            type=17,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='status_message',
            full_name='spotify.MercuryReply.status_message',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='cache_policy',
            full_name='spotify.MercuryReply.cache_policy',
            index=2,
            number=3,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=1,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='ttl',
            full_name='spotify.MercuryReply.ttl',
            index=3,
            number=4,
            type=17,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='etag',
            full_name='spotify.MercuryReply.etag',
            index=4,
            number=5,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='content_type',
            full_name='spotify.MercuryReply.content_type',
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='body',
            full_name='spotify.MercuryReply.body',
            index=6,
            number=7,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[
        _MERCURYREPLY_CACHEPOLICY,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=238,
    serialized_end=483,
)

_HEADER = _descriptor.Descriptor(
    name='Header',
    full_name='spotify.Header',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='uri',
            full_name='spotify.Header.uri',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='content_type',
            full_name='spotify.Header.content_type',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='method',
            full_name='spotify.Header.method',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='status_code',
            full_name='spotify.Header.status_code',
            index=3,
            number=4,
            type=17,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='user_fields',
            full_name='spotify.Header.user_fields',
            index=4,
            number=6,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=485,
    serialized_end=606,
)

_USERFIELD = _descriptor.Descriptor(
    name='UserField',
    full_name='spotify.UserField',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='spotify.UserField.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='spotify.UserField.value',
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[],
    serialized_start=608,
    serialized_end=647,
)

_MERCURYMULTIGETREQUEST.fields_by_name[
    'request'].message_type = _MERCURYREQUEST
_MERCURYMULTIGETREPLY.fields_by_name['reply'].message_type = _MERCURYREPLY
_MERCURYREPLY.fields_by_name[
    'cache_policy'].enum_type = _MERCURYREPLY_CACHEPOLICY
_MERCURYREPLY_CACHEPOLICY.containing_type = _MERCURYREPLY
_HEADER.fields_by_name['user_fields'].message_type = _USERFIELD
DESCRIPTOR.message_types_by_name[
    'MercuryMultiGetRequest'] = _MERCURYMULTIGETREQUEST
DESCRIPTOR.message_types_by_name[
    'MercuryMultiGetReply'] = _MERCURYMULTIGETREPLY
DESCRIPTOR.message_types_by_name['MercuryRequest'] = _MERCURYREQUEST
DESCRIPTOR.message_types_by_name['MercuryReply'] = _MERCURYREPLY
DESCRIPTOR.message_types_by_name['Header'] = _HEADER
DESCRIPTOR.message_types_by_name['UserField'] = _USERFIELD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MercuryMultiGetRequest = _reflection.GeneratedProtocolMessageType(
    'MercuryMultiGetRequest',
    (_message.Message, ),
    {
        'DESCRIPTOR': _MERCURYMULTIGETREQUEST,
        '__module__': 'mercury_pb2'
        # @@protoc_insertion_point(class_scope:spotify.MercuryMultiGetRequest)
    })
_sym_db.RegisterMessage(MercuryMultiGetRequest)

MercuryMultiGetReply = _reflection.GeneratedProtocolMessageType(
    'MercuryMultiGetReply',
    (_message.Message, ),
    {
        'DESCRIPTOR': _MERCURYMULTIGETREPLY,
        '__module__': 'mercury_pb2'
        # @@protoc_insertion_point(class_scope:spotify.MercuryMultiGetReply)
    })
_sym_db.RegisterMessage(MercuryMultiGetReply)

MercuryRequest = _reflection.GeneratedProtocolMessageType(
    'MercuryRequest',
    (_message.Message, ),
    {
        'DESCRIPTOR': _MERCURYREQUEST,
        '__module__': 'mercury_pb2'
        # @@protoc_insertion_point(class_scope:spotify.MercuryRequest)
    })
_sym_db.RegisterMessage(MercuryRequest)

MercuryReply = _reflection.GeneratedProtocolMessageType(
    'MercuryReply',
    (_message.Message, ),
    {
        'DESCRIPTOR': _MERCURYREPLY,
        '__module__': 'mercury_pb2'
        # @@protoc_insertion_point(class_scope:spotify.MercuryReply)
    })
_sym_db.RegisterMessage(MercuryReply)

Header = _reflection.GeneratedProtocolMessageType(
    'Header',
    (_message.Message, ),
    {
        'DESCRIPTOR': _HEADER,
        '__module__': 'mercury_pb2'
        # @@protoc_insertion_point(class_scope:spotify.Header)
    })
_sym_db.RegisterMessage(Header)

UserField = _reflection.GeneratedProtocolMessageType(
    'UserField',
    (_message.Message, ),
    {
        'DESCRIPTOR': _USERFIELD,
        '__module__': 'mercury_pb2'
        # @@protoc_insertion_point(class_scope:spotify.UserField)
    })
_sym_db.RegisterMessage(UserField)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: spotify/login5/v3/credentials/credentials.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name="spotify/login5/v3/credentials/credentials.proto",
    package="spotify.login5.v3.credentials",
    syntax="proto3",
    serialized_options=b"\n\024com.spotify.login5v3",
    create_key=_descriptor._internal_create_key,
    serialized_pb=
    b'\n/spotify/login5/v3/credentials/credentials.proto\x12\x1dspotify.login5.v3.credentials"2\n\x10StoredCredential\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c"9\n\x08Password\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\x12\x0f\n\x07padding\x18\x03 \x01(\x0c";\n\x13\x46\x61\x63\x65\x62ookAccessToken\x12\x0e\n\x06\x66\x62_uid\x18\x01 \x01(\t\x12\x14\n\x0c\x61\x63\x63\x65ss_token\x18\x02 \x01(\t"\x1d\n\x0cOneTimeToken\x12\r\n\x05token\x18\x01 \x01(\t"|\n\x15ParentChildCredential\x12\x10\n\x08\x63hild_id\x18\x01 \x01(\t\x12Q\n\x18parent_stored_credential\x18\x02 \x01(\x0b\x32/.spotify.login5.v3.credentials.StoredCredential"S\n\x15\x41ppleSignInCredential\x12\x11\n\tauth_code\x18\x01 \x01(\t\x12\x14\n\x0credirect_uri\x18\x02 \x01(\t\x12\x11\n\tbundle_id\x18\x03 \x01(\tB\x16\n\x14\x63om.spotify.login5v3b\x06proto3',
)

_STOREDCREDENTIAL = _descriptor.Descriptor(
    name="StoredCredential",
    full_name="spotify.login5.v3.credentials.StoredCredential",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="username",
            full_name="spotify.login5.v3.credentials.StoredCredential.username",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="spotify.login5.v3.credentials.StoredCredential.data",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=82,
    serialized_end=132,
)

_PASSWORD = _descriptor.Descriptor(
    name="Password",
    full_name="spotify.login5.v3.credentials.Password",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="spotify.login5.v3.credentials.Password.id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="password",
            full_name="spotify.login5.v3.credentials.Password.password",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="padding",
            full_name="spotify.login5.v3.credentials.Password.padding",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=134,
    serialized_end=191,
)

_FACEBOOKACCESSTOKEN = _descriptor.Descriptor(
    name="FacebookAccessToken",
    full_name="spotify.login5.v3.credentials.FacebookAccessToken",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="fb_uid",
            full_name=
            "spotify.login5.v3.credentials.FacebookAccessToken.fb_uid",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="access_token",
            full_name=
            "spotify.login5.v3.credentials.FacebookAccessToken.access_token",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=193,
    serialized_end=252,
)

_ONETIMETOKEN = _descriptor.Descriptor(
    name="OneTimeToken",
    full_name="spotify.login5.v3.credentials.OneTimeToken",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="token",
            full_name="spotify.login5.v3.credentials.OneTimeToken.token",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=254,
    serialized_end=283,
)

_PARENTCHILDCREDENTIAL = _descriptor.Descriptor(
    name="ParentChildCredential",
    full_name="spotify.login5.v3.credentials.ParentChildCredential",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="child_id",
            full_name=
            "spotify.login5.v3.credentials.ParentChildCredential.child_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="parent_stored_credential",
            full_name=
            "spotify.login5.v3.credentials.ParentChildCredential.parent_stored_credential",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=285,
    serialized_end=409,
)

_APPLESIGNINCREDENTIAL = _descriptor.Descriptor(
    name="AppleSignInCredential",
    full_name="spotify.login5.v3.credentials.AppleSignInCredential",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="auth_code",
            full_name=
            "spotify.login5.v3.credentials.AppleSignInCredential.auth_code",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="redirect_uri",
            full_name=
            "spotify.login5.v3.credentials.AppleSignInCredential.redirect_uri",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="bundle_id",
            full_name=
            "spotify.login5.v3.credentials.AppleSignInCredential.bundle_id",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=411,
    serialized_end=494,
)

_PARENTCHILDCREDENTIAL.fields_by_name[
    "parent_stored_credential"].message_type = _STOREDCREDENTIAL
DESCRIPTOR.message_types_by_name["StoredCredential"] = _STOREDCREDENTIAL
DESCRIPTOR.message_types_by_name["Password"] = _PASSWORD
DESCRIPTOR.message_types_by_name["FacebookAccessToken"] = _FACEBOOKACCESSTOKEN
DESCRIPTOR.message_types_by_name["OneTimeToken"] = _ONETIMETOKEN
DESCRIPTOR.message_types_by_name[
    "ParentChildCredential"] = _PARENTCHILDCREDENTIAL
DESCRIPTOR.message_types_by_name[
    "AppleSignInCredential"] = _APPLESIGNINCREDENTIAL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StoredCredential = _reflection.GeneratedProtocolMessageType(
    "StoredCredential",
    (_message.Message, ),
    {
        "DESCRIPTOR": _STOREDCREDENTIAL,
        "__module__": "spotify.login5.v3.credentials.credentials_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.credentials.StoredCredential)
    },
)
_sym_db.RegisterMessage(StoredCredential)

Password = _reflection.GeneratedProtocolMessageType(
    "Password",
    (_message.Message, ),
    {
        "DESCRIPTOR": _PASSWORD,
        "__module__": "spotify.login5.v3.credentials.credentials_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.credentials.Password)
    },
)
_sym_db.RegisterMessage(Password)

FacebookAccessToken = _reflection.GeneratedProtocolMessageType(
    "FacebookAccessToken",
    (_message.Message, ),
    {
        "DESCRIPTOR": _FACEBOOKACCESSTOKEN,
        "__module__": "spotify.login5.v3.credentials.credentials_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.credentials.FacebookAccessToken)
    },
)
_sym_db.RegisterMessage(FacebookAccessToken)

OneTimeToken = _reflection.GeneratedProtocolMessageType(
    "OneTimeToken",
    (_message.Message, ),
    {
        "DESCRIPTOR": _ONETIMETOKEN,
        "__module__": "spotify.login5.v3.credentials.credentials_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.credentials.OneTimeToken)
    },
)
_sym_db.RegisterMessage(OneTimeToken)

ParentChildCredential = _reflection.GeneratedProtocolMessageType(
    "ParentChildCredential",
    (_message.Message, ),
    {
        "DESCRIPTOR": _PARENTCHILDCREDENTIAL,
        "__module__": "spotify.login5.v3.credentials.credentials_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.credentials.ParentChildCredential)
    },
)
_sym_db.RegisterMessage(ParentChildCredential)

AppleSignInCredential = _reflection.GeneratedProtocolMessageType(
    "AppleSignInCredential",
    (_message.Message, ),
    {
        "DESCRIPTOR": _APPLESIGNINCREDENTIAL,
        "__module__": "spotify.login5.v3.credentials.credentials_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.credentials.AppleSignInCredential)
    },
)
_sym_db.RegisterMessage(AppleSignInCredential)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
from Cryptodome.Cipher import AES
from Cryptodome.Hash import HMAC, SHA1
from Cryptodome.Util import Counter
from librespot import util, Version
from librespot.core import Session
from librespot.crypto import DiffieHellman
from librespot.proto import Connect_pb2 as Connect
from librespot.structure import Closeable, Runnable, SessionListener
import base64
import concurrent.futures
import copy
import io
import json
import logging
import random
import socket
import threading
import typing
import urllib.parse
import zeroconf


class ZeroconfServer(Closeable):
    logger = logging.getLogger("Librespot:ZeroconfServer")
    logger.propagate = False
    
    service = "_spotify-connect._tcp.local."
    __connecting_username: typing.Union[str, None] = None
    __connection_lock = threading.Condition()
    __default_get_info_fields = {
        "status": 101,
        "statusString": "OK",
        "spotifyError": 0,
        "version": "2.9.0",
        "libraryVersion": Version.version_name,
        "accountReq": "FREE",
        "brandDisplayName": "Spotizerr",
        "modelDisplayName": "librespot-spotizerr",
        "voiceSupport": "NO",
        "availability": "1",
        "productID": 0,
        "tokenType": "default",
        "groupStatus": "NONE",
        "resolverVersion": "1",
        "scope": "streaming,client-authorization-universal",
        "clientID": "65b708073fc0480ea92a077233ca87bd"
    }
    __default_successful_add_user = {
        "status": 101,
        "spotifyError": 0,
        "statusString": "OK",
    }
    __eol = b"\r\n"
    __max_port = 65536
    __min_port = 1024
    __runner: HttpRunner
    __service_info: zeroconf.ServiceInfo
    __session: typing.Union[Session, None] = None
    __session_listeners: typing.List[SessionListener] = []
    __zeroconf: zeroconf.Zeroconf

    def __init__(self, inner: Inner, listen_port):
        self.__inner = inner
        self.__keys = DiffieHellman()
        if listen_port == -1:
            listen_port = random.randint(self.__min_port + 1, self.__max_port)
        self.__runner = ZeroconfServer.HttpRunner(self, listen_port)
        threading.Thread(target=self.__runner.run,
                         name="zeroconf-http-server",
                         daemon=True).start()
                         
        self.__zeroconf = zeroconf.Zeroconf()
        
        advertised_ip_str = self._get_local_ip()
        
        server_hostname = socket.gethostname()
        if not server_hostname or server_hostname == "localhost":
            self.logger.warning(
                f"Machine hostname is '{server_hostname}', which is not ideal for mDNS. "
                f"Consider setting a unique hostname for this machine. "
                f"Using device name '{inner.device_name}' as part of the service instance name, "
                f"but relying on zeroconf library to handle server resolution for IP {advertised_ip_str}."
            )

        service_addresses = None
        if advertised_ip_str != "0.0.0.0":
            try:
                service_addresses = [socket.inet_aton(advertised_ip_str)]
            except socket.error: # Catches errors like invalid IP string format
                self.logger.error(f"Failed to convert IP string '{advertised_ip_str}' to packed address. Zeroconf will attempt to determine addresses from hostname.")
        
        self.__service_info = zeroconf.ServiceInfo(
            ZeroconfServer.service,  # type, e.g., "_spotify-connect._tcp.local."
            f"{inner.device_name}.{ZeroconfServer.service}",  # name, e.g., "MyDevice._spotify-connect._tcp.local."
            listen_port,
            0,  # weight
            0,  # priority
            {   # properties
                "CPath": "/",
                "VERSION": "1.0",
                "STACK": "SP",
            },
            server=f"{server_hostname}.local.", # server FQDN
            addresses=service_addresses # Pass resolved IP, or None if 0.0.0.0 or conversion failed
        )
        
        self.__zeroconf.register_service(self.__service_info)

    def _get_local_ip(self) -> str:
        """Tries to determine a non-loopback local IP address for network advertisement."""
        s = None
        ip_address = "0.0.0.0" # Default to all interfaces if specific IP cannot be found
        try:
            # Attempt to connect to an external address to find the appropriate local IP
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.settimeout(0.1) # Short timeout for the connect attempt
            s.connect(("8.8.8.8", 80)) # Google's public DNS server
            ip_address = s.getsockname()[0]
            self.logger.info(f"Determined local IP via connect trick: {ip_address}")
        except OSError as e: # Catches socket errors like [Errno 101] Network is unreachable
            self.logger.warning(f"Could not determine local IP via connect trick ({e}). Trying socket.gethostname().")
            try:
                hostname = socket.gethostname()
                # gethostbyname can return 127.0.0.1 if hostname resolves to it
                ip_address = socket.gethostbyname(hostname)
                self.logger.info(f"IP from socket.gethostname('{hostname}'): {ip_address}")
            except socket.gaierror:
                self.logger.error(
                    f"socket.gaierror resolving hostname '{socket.gethostname()}'. Falling back to 0.0.0.0."
                )
                ip_address = "0.0.0.0"
        finally:
            if s:
                s.close()

        # If the IP is loopback (127.x.x.x) or still 0.0.0.0, try to find a better one from all interfaces
        if ip_address.startswith("127.") or ip_address == "0.0.0.0":
            self.logger.warning(
                f"Current IP ('{ip_address}') is loopback or 0.0.0.0. Attempting to find a non-loopback IP from host interfaces."
            )
            try:
                current_hostname = socket.gethostname()
                all_ips_info = socket.gethostbyname_ex(current_hostname)
                # all_ips_info is a tuple: (hostname, aliaslist, ipaddrlist)
                non_loopback_ips = [ip for ip in all_ips_info[2] if not ip.startswith("127.")]
                
                if non_loopback_ips:
                    ip_address = non_loopback_ips[0] # Pick the first non-loopback IP
                    self.logger.info(f"Found non-loopback IP from host interfaces ('{current_hostname}'): {ip_address}")
                else:
                    self.logger.warning(
                        f"No non-loopback IPs found for hostname '{current_hostname}'. Retaining '{ip_address}'."
                    )
                    # If ip_address was 0.0.0.0, it remains so. If it was 127.0.0.1, it remains so.
            except socket.gaierror:
                self.logger.error(
                    f"socket.gaierror during fallback IP search for hostname '{socket.gethostname()}'. Retaining '{ip_address}'."
                )
        
        if ip_address == "0.0.0.0":
            self.logger.warning("Failed to determine a specific non-loopback IP. Zeroconf will attempt to use all available interfaces.")
        elif ip_address.startswith("127."):
             self.logger.warning(f"Final IP for advertisement is loopback ('{ip_address}'). Service discovery may not work correctly on the network.")
        else:
            self.logger.info(f"Using IP address for Zeroconf advertisement: {ip_address}")
            
        return ip_address

    def add_session_listener(self, listener: ZeroconfServer):
        self.__session_listeners.append(listener)

    def close(self) -> None:
        self.__zeroconf.close()
        self.__runner.close()

    def close_session(self) -> None:
        if self.__session is None:
            return
        for session_listener in self.__session_listeners:
            session_listener.session_closing(self.__session)
        self.__session.close()
        self.__session = None

    def handle_add_user(self, __socket: socket.socket, params: dict[str, str],
                        http_version: str) -> None:
        username = params.get("userName")
        if not username:
            self.logger.error("Missing userName!")
            return
        blob_str = params.get("blob")
        if not blob_str:
            self.logger.error("Missing blob!")
            return
        client_key_str = params.get("clientKey")
        if not client_key_str:
            self.logger.error("Missing clientKey!")
        with self.__connection_lock:
            if username == self.__connecting_username:
                self.logger.info(
                    "{} is already trying to connect.".format(username))
                __socket.send(http_version.encode())
                __socket.send(b" 403 Forbidden")
                __socket.send(self.__eol)
                __socket.send(self.__eol)
                return
        shared_key = util.int_to_bytes(
            self.__keys.compute_shared_key(
                base64.b64decode(client_key_str.encode())))
        blob_bytes = base64.b64decode(blob_str)
        iv = blob_bytes[:16]
        encrypted = blob_bytes[16:len(blob_bytes) - 20]
        checksum = blob_bytes[len(blob_bytes) - 20:]
        sha1 = SHA1.new()
        sha1.update(shared_key)
        base_key = sha1.digest()[:16]
        hmac = HMAC.new(base_key, digestmod=SHA1)
        hmac.update(b"checksum")
        checksum_key = hmac.digest()
        hmac = HMAC.new(base_key, digestmod=SHA1)
        hmac.update(b"encryption")
        encryption_key = hmac.digest()
        hmac = HMAC.new(checksum_key, digestmod=SHA1)
        hmac.update(encrypted)
        mac = hmac.digest()
        if mac != checksum:
            self.logger.error("Mac and checksum don't match!")
            __socket.send(http_version.encode())
            __socket.send(b" 400 Bad Request")
            __socket.send(self.__eol)
            __socket.send(self.__eol)
            return
        aes = AES.new(encryption_key[:16],
                      AES.MODE_CTR,
                      counter=Counter.new(128,
                                          initial_value=int.from_bytes(
                                              iv, "big")))
        decrypted = aes.decrypt(encrypted)
        self.close_session()
        with self.__connection_lock:
            self.__connecting_username = username
        self.logger.info("Accepted new user from {}. [deviceId: {}]".format(
            params.get("deviceName"), self.__inner.device_id))
        response = json.dumps(self.__default_successful_add_user)
        __socket.send(http_version.encode())
        __socket.send(b" 200 OK")
        __socket.send(self.__eol)
        __socket.send(b"Content-Length: ")
        __socket.send(str(len(response)).encode())
        __socket.send(self.__eol)
        __socket.send(self.__eol)
        __socket.send(response.encode())
        self.__session = Session.Builder(self.__inner.conf) \
            .set_device_id(self.__inner.device_id) \
            .set_device_name(self.__inner.device_name) \
            .set_device_type(self.__inner.device_type) \
            .set_preferred_locale(self.__inner.preferred_locale) \
            .blob(username, decrypted) \
            .create()
        with self.__connection_lock:
            self.__connecting_username = None
        for session_listener in self.__session_listeners:
            session_listener.session_changed(self.__session)

    def handle_get_info(self, __socket: socket.socket,
                        http_version: str) -> None:
        info = copy.deepcopy(self.__default_get_info_fields)
        info["deviceID"] = self.__inner.device_id
        info["remoteName"] = self.__inner.device_name
        info["publicKey"] = base64.b64encode(
            self.__keys.public_key_bytes()).decode()
        device_type_name = Connect.DeviceType.Name(self.__inner.device_type)
        info["deviceType"] = device_type_name.title()
        with self.__connection_lock:
            info[
                "activeUser"] = self.__connecting_username if self.__connecting_username is not None else self.__session.username(
                ) if self.has_valid_session() else ""
        __socket.send(http_version.encode())
        __socket.send(b" 200 OK")
        __socket.send(self.__eol)
        __socket.send(b"Content-Type: application/json")
        __socket.send(self.__eol)
        __socket.send(self.__eol)
        __socket.send(json.dumps(info).encode())

    def has_valid_session(self) -> bool:
        valid = self.__session and self.__session.is_valid()
        if not valid:
            self.__session = None
        return valid

    def parse_path(self, path: str) -> dict[str, str]:
        url = "https://host" + path
        parsed = {}
        params = urllib.parse.parse_qs(urllib.parse.urlparse(url).query)
        for key, values in params.items():
            for value in values:
                parsed[key] = value
        return parsed

    def remove_session_listener(self, listener: SessionListener):
        self.__session_listeners.remove(listener)

    class Builder(Session.Builder):
        listen_port: int = -1

        def set_listen_port(self, listen_port: int):
            self.listen_port = listen_port
            return self

        def create(self) -> ZeroconfServer:
            return ZeroconfServer(
                ZeroconfServer.Inner(self.device_type, self.device_name,
                                     self.device_id, self.preferred_locale,
                                     self.conf), self.listen_port)

    class HttpRunner(Closeable, Runnable):
        __should_stop = False
        __socket: socket.socket
        __worker = concurrent.futures.ThreadPoolExecutor()
        __zeroconf_server: ZeroconfServer

        def __init__(self, zeroconf_server: ZeroconfServer, port: int):
            self.__socket = socket.socket()
            self.__socket.bind((".".join(["0"] * 4), port))
            self.__socket.listen(5)
            self.__zeroconf_server = zeroconf_server
            self.__zeroconf_server.logger.info(
                "Zeroconf HTTP server started successfully on port {}!".format(
                    port))

        def close(self) -> None:
            pass

        def run(self):
            while not self.__should_stop:
                __socket, address = self.__socket.accept()

                def anonymous():
                    self.__handle(__socket)
                    __socket.close()

                self.__worker.submit(anonymous)

        def __handle(self, __socket: socket.socket) -> None:
            request = io.BytesIO(__socket.recv(1024 * 1024))
            request_line = request.readline().strip().split(b" ")
            if len(request_line) != 3:
                self.__zeroconf_server.logger.warning(
                    "Unexpected request line: {}".format(request_line))
            method = request_line[0].decode()
            path = request_line[1].decode()
            http_version = request_line[2].decode()
            headers = {}
            while True:
                header = request.readline().strip()
                if not header:
                    break
                split = header.split(b":")
                headers[split[0].decode()] = split[1].strip().decode()
            if not self.__zeroconf_server.has_valid_session():
                self.__zeroconf_server.logger.debug(
                    "Handling request: {}, {}, {}, headers: {}".format(
                        method, path, http_version, headers))
            params = {}
            if method == "POST":
                content_type = headers.get("Content-Type")
                if content_type != "application/x-www-form-urlencoded":
                    self.__zeroconf_server.logger.error(
                        "Bad Content-Type: {}".format(content_type))
                    return
                content_length_str = headers.get("Content-Length")
                if content_length_str is None:
                    self.__zeroconf_server.logger.error(
                        "Missing Content-Length header!")
                    return
                content_length = int(content_length_str)
                body = request.read(content_length).decode()
                pairs = body.split("&")
                for pair in pairs:
                    split = pair.split("=")
                    params[urllib.parse.unquote(
                        split[0])] = urllib.parse.unquote(split[1])
            else:
                params = self.__zeroconf_server.parse_path(path)
            action = params.get("action")
            if action is None:
                self.__zeroconf_server.logger.debug(
                    "Request is missing action.")
                return
            self.handle_request(__socket, http_version, action, params)

        def handle_request(self, __socket: socket.socket, http_version: str,
                           action: str, params: dict[str, str]) -> None:
            if action == "addUser":
                if params is None:
                    raise RuntimeError
                self.__zeroconf_server.handle_add_user(__socket, params,
                                                       http_version)
            elif action == "getInfo":
                self.__zeroconf_server.handle_get_info(__socket, http_version)
            else:
                self.__zeroconf_server.logger.warning(
                    "Unknown action: {}".format(action))

    class Inner:
        conf: typing.Final[Session.Configuration]
        device_name: typing.Final[str]
        device_id: typing.Final[str]
        device_type: typing.Final[Connect.DeviceType]
        preferred_locale: typing.Final[str]

        def __init__(self, device_type: Connect.DeviceType, device_name: str,
                     device_id: str, preferred_locale: str,
                     conf: Session.Configuration):
            self.conf = conf
            self.device_name = device_name
            self.device_id = util.random_hex_string(
                40).lower() if not device_id else device_id
            self.device_type = device_type
            self.preferred_locale = preferred_locale


from __future__ import annotations
from librespot.crypto import DiffieHellman
from librespot.proto.Keyexchange_pb2 import BuildInfo, Platform, Product, ProductFlags
from librespot.structure import Closeable, Runnable
import platform


class Version:
    version_name = "0.3.0"

    @staticmethod
    def platform() -> Platform:
        if platform.system() == "Windows":
            return Platform.PLATFORM_WIN32_X86
        if platform.system() == "Darwin":
            return Platform.PLATFORM_OSX_X86
        return Platform.PLATFORM_LINUX_X86

    @staticmethod
    def version_string():
        return "librespot-spotizerr " + Version.version_name

    @staticmethod
    def system_info_string():
        return Version.version_string() + \
               "; Python " + platform.python_version() + \
               "; " + platform.system()

    @staticmethod
    def standard_build_info() -> BuildInfo:
        return BuildInfo(product=Product.PRODUCT_CLIENT,
                         product_flags=[ProductFlags.PRODUCT_FLAG_NONE],
                         platform=Version.platform(),
                         version=117300517)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: spotify/login5/v3/login5.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import enum_type_wrapper
from librespot.proto.spotify.login5.v3 import \
    ClientInfo_pb2 as spotify_dot_login5_dot_v3_dot_client__info__pb2
from librespot.proto.spotify.login5.v3 import \
    UserInfo_pb2 as spotify_dot_login5_dot_v3_dot_user__info__pb2
from librespot.proto.spotify.login5.v3.challenges import \
    Code_pb2 as spotify_dot_login5_dot_v3_dot_challenges_dot_code__pb2
from librespot.proto.spotify.login5.v3.challenges import \
    Hashcash_pb2 as spotify_dot_login5_dot_v3_dot_challenges_dot_hashcash__pb2
from librespot.proto.spotify.login5.v3.credentials import \
    Credentials_pb2 as \
    spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2
from librespot.proto.spotify.login5.v3.identifiers import \
    Identifiers as \
    spotify_dot_login5_dot_v3_dot_identifiers_dot_identifiers__pb2

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name="spotify/login5/v3/login5.proto",
    package="spotify.login5.v3",
    syntax="proto3",
    serialized_options=b"\n\024com.spotify.login5v3",
    create_key=_descriptor._internal_create_key,
    serialized_pb=
    b'\n\x1espotify/login5/v3/login5.proto\x12\x11spotify.login5.v3\x1a#spotify/login5/v3/client_info.proto\x1a!spotify/login5/v3/user_info.proto\x1a\'spotify/login5/v3/challenges/code.proto\x1a+spotify/login5/v3/challenges/hashcash.proto\x1a/spotify/login5/v3/credentials/credentials.proto\x1a/spotify/login5/v3/identifiers/identifiers.proto">\n\nChallenges\x12\x30\n\nchallenges\x18\x01 \x03(\x0b\x32\x1c.spotify.login5.v3.Challenge"\x89\x01\n\tChallenge\x12\x41\n\x08hashcash\x18\x01 \x01(\x0b\x32/.spotify.login5.v3.challenges.HashcashChallenge\x12\x39\n\x04\x63ode\x18\x02 \x01(\x0b\x32+.spotify.login5.v3.challenges.CodeChallenge"M\n\x12\x43hallengeSolutions\x12\x37\n\tsolutions\x18\x01 \x03(\x0b\x32$.spotify.login5.v3.ChallengeSolution"\x8f\x01\n\x11\x43hallengeSolution\x12@\n\x08hashcash\x18\x01 \x01(\x0b\x32..spotify.login5.v3.challenges.HashcashSolution\x12\x38\n\x04\x63ode\x18\x02 \x01(\x0b\x32*.spotify.login5.v3.challenges.CodeSolution"\xad\x05\n\x0cLoginRequest\x12\x32\n\x0b\x63lient_info\x18\x01 \x01(\x0b\x32\x1d.spotify.login5.v3.ClientInfo\x12\x15\n\rlogin_context\x18\x02 \x01(\x0c\x12\x42\n\x13\x63hallenge_solutions\x18\x03 \x01(\x0b\x32%.spotify.login5.v3.ChallengeSolutions\x12J\n\x11stored_credential\x18\x64 \x01(\x0b\x32/.spotify.login5.v3.credentials.StoredCredential\x12\x39\n\x08password\x18\x65 \x01(\x0b\x32\'.spotify.login5.v3.credentials.Password\x12Q\n\x15\x66\x61\x63\x65\x62ook_access_token\x18\x66 \x01(\x0b\x32\x32.spotify.login5.v3.credentials.FacebookAccessToken\x12@\n\x0cphone_number\x18g \x01(\x0b\x32*.spotify.login5.v3.identifiers.PhoneNumber\x12\x43\n\x0eone_time_token\x18h \x01(\x0b\x32+.spotify.login5.v3.credentials.OneTimeToken\x12U\n\x17parent_child_credential\x18i \x01(\x0b\x32\x34.spotify.login5.v3.credentials.ParentChildCredential\x12V\n\x18\x61pple_sign_in_credential\x18j \x01(\x0b\x32\x34.spotify.login5.v3.credentials.AppleSignInCredential"m\n\x07LoginOk\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x14\n\x0c\x61\x63\x63\x65ss_token\x18\x02 \x01(\t\x12\x19\n\x11stored_credential\x18\x03 \x01(\x0c\x12\x1f\n\x17\x61\x63\x63\x65ss_token_expires_in\x18\x04 \x01(\x05"\xf8\x02\n\rLoginResponse\x12&\n\x02ok\x18\x01 \x01(\x0b\x32\x1a.spotify.login5.v3.LoginOk\x12,\n\x05\x65rror\x18\x02 \x01(\x0e\x32\x1d.spotify.login5.v3.LoginError\x12\x31\n\nchallenges\x18\x03 \x01(\x0b\x32\x1d.spotify.login5.v3.Challenges\x12;\n\x08warnings\x18\x04 \x03(\x0e\x32).spotify.login5.v3.LoginResponse.Warnings\x12\x15\n\rlogin_context\x18\x05 \x01(\x0c\x12\x18\n\x10identifier_token\x18\x06 \x01(\t\x12.\n\tuser_info\x18\x07 \x01(\x0b\x32\x1b.spotify.login5.v3.UserInfo"@\n\x08Warnings\x12\x13\n\x0fUNKNOWN_WARNING\x10\x00\x12\x1f\n\x1b\x44\x45PRECATED_PROTOCOL_VERSION\x10\x01*\xd3\x01\n\nLoginError\x12\x11\n\rUNKNOWN_ERROR\x10\x00\x12\x17\n\x13INVALID_CREDENTIALS\x10\x01\x12\x0f\n\x0b\x42\x41\x44_REQUEST\x10\x02\x12\x1e\n\x1aUNSUPPORTED_LOGIN_PROTOCOL\x10\x03\x12\x0b\n\x07TIMEOUT\x10\x04\x12\x16\n\x12UNKNOWN_IDENTIFIER\x10\x05\x12\x15\n\x11TOO_MANY_ATTEMPTS\x10\x06\x12\x17\n\x13INVALID_PHONENUMBER\x10\x07\x12\x13\n\x0fTRY_AGAIN_LATER\x10\x08\x42\x16\n\x14\x63om.spotify.login5v3b\x06proto3',
    dependencies=[
        spotify_dot_login5_dot_v3_dot_client__info__pb2.DESCRIPTOR,
        spotify_dot_login5_dot_v3_dot_user__info__pb2.DESCRIPTOR,
        spotify_dot_login5_dot_v3_dot_challenges_dot_code__pb2.DESCRIPTOR,
        spotify_dot_login5_dot_v3_dot_challenges_dot_hashcash__pb2.DESCRIPTOR,
        spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2.
        DESCRIPTOR,
        spotify_dot_login5_dot_v3_dot_identifiers_dot_identifiers__pb2.
        DESCRIPTOR,
    ],
)

_LOGINERROR = _descriptor.EnumDescriptor(
    name="LoginError",
    full_name="spotify.login5.v3.LoginError",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNKNOWN_ERROR",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="INVALID_CREDENTIALS",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="BAD_REQUEST",
            index=2,
            number=2,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="UNSUPPORTED_LOGIN_PROTOCOL",
            index=3,
            number=3,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TIMEOUT",
            index=4,
            number=4,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="UNKNOWN_IDENTIFIER",
            index=5,
            number=5,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TOO_MANY_ATTEMPTS",
            index=6,
            number=6,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="INVALID_PHONENUMBER",
            index=7,
            number=7,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="TRY_AGAIN_LATER",
            index=8,
            number=8,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=1917,
    serialized_end=2128,
)
_sym_db.RegisterEnumDescriptor(_LOGINERROR)

LoginError = enum_type_wrapper.EnumTypeWrapper(_LOGINERROR)
UNKNOWN_ERROR = 0
INVALID_CREDENTIALS = 1
BAD_REQUEST = 2
UNSUPPORTED_LOGIN_PROTOCOL = 3
TIMEOUT = 4
UNKNOWN_IDENTIFIER = 5
TOO_MANY_ATTEMPTS = 6
INVALID_PHONENUMBER = 7
TRY_AGAIN_LATER = 8

_LOGINRESPONSE_WARNINGS = _descriptor.EnumDescriptor(
    name="Warnings",
    full_name="spotify.login5.v3.LoginResponse.Warnings",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="UNKNOWN_WARNING",
            index=0,
            number=0,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.EnumValueDescriptor(
            name="DEPRECATED_PROTOCOL_VERSION",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=1850,
    serialized_end=1914,
)
_sym_db.RegisterEnumDescriptor(_LOGINRESPONSE_WARNINGS)

_CHALLENGES = _descriptor.Descriptor(
    name="Challenges",
    full_name="spotify.login5.v3.Challenges",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="challenges",
            full_name="spotify.login5.v3.Challenges.challenges",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=309,
    serialized_end=371,
)

_CHALLENGE = _descriptor.Descriptor(
    name="Challenge",
    full_name="spotify.login5.v3.Challenge",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="hashcash",
            full_name="spotify.login5.v3.Challenge.hashcash",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="code",
            full_name="spotify.login5.v3.Challenge.code",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=374,
    serialized_end=511,
)

_CHALLENGESOLUTIONS = _descriptor.Descriptor(
    name="ChallengeSolutions",
    full_name="spotify.login5.v3.ChallengeSolutions",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="solutions",
            full_name="spotify.login5.v3.ChallengeSolutions.solutions",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=513,
    serialized_end=590,
)

_CHALLENGESOLUTION = _descriptor.Descriptor(
    name="ChallengeSolution",
    full_name="spotify.login5.v3.ChallengeSolution",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="hashcash",
            full_name="spotify.login5.v3.ChallengeSolution.hashcash",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="code",
            full_name="spotify.login5.v3.ChallengeSolution.code",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=593,
    serialized_end=736,
)

_LOGINREQUEST = _descriptor.Descriptor(
    name="LoginRequest",
    full_name="spotify.login5.v3.LoginRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="client_info",
            full_name="spotify.login5.v3.LoginRequest.client_info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="login_context",
            full_name="spotify.login5.v3.LoginRequest.login_context",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="challenge_solutions",
            full_name="spotify.login5.v3.LoginRequest.challenge_solutions",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="stored_credential",
            full_name="spotify.login5.v3.LoginRequest.stored_credential",
            index=3,
            number=100,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="password",
            full_name="spotify.login5.v3.LoginRequest.password",
            index=4,
            number=101,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="facebook_access_token",
            full_name="spotify.login5.v3.LoginRequest.facebook_access_token",
            index=5,
            number=102,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="phone_number",
            full_name="spotify.login5.v3.LoginRequest.phone_number",
            index=6,
            number=103,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="one_time_token",
            full_name="spotify.login5.v3.LoginRequest.one_time_token",
            index=7,
            number=104,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="parent_child_credential",
            full_name="spotify.login5.v3.LoginRequest.parent_child_credential",
            index=8,
            number=105,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="apple_sign_in_credential",
            full_name="spotify.login5.v3.LoginRequest.apple_sign_in_credential",
            index=9,
            number=106,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=739,
    serialized_end=1424,
)

_LOGINOK = _descriptor.Descriptor(
    name="LoginOk",
    full_name="spotify.login5.v3.LoginOk",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="username",
            full_name="spotify.login5.v3.LoginOk.username",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="access_token",
            full_name="spotify.login5.v3.LoginOk.access_token",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="stored_credential",
            full_name="spotify.login5.v3.LoginOk.stored_credential",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="access_token_expires_in",
            full_name="spotify.login5.v3.LoginOk.access_token_expires_in",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1426,
    serialized_end=1535,
)

_LOGINRESPONSE = _descriptor.Descriptor(
    name="LoginResponse",
    full_name="spotify.login5.v3.LoginResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="ok",
            full_name="spotify.login5.v3.LoginResponse.ok",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="error",
            full_name="spotify.login5.v3.LoginResponse.error",
            index=1,
            number=2,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="challenges",
            full_name="spotify.login5.v3.LoginResponse.challenges",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="warnings",
            full_name="spotify.login5.v3.LoginResponse.warnings",
            index=3,
            number=4,
            type=14,
            cpp_type=8,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="login_context",
            full_name="spotify.login5.v3.LoginResponse.login_context",
            index=4,
            number=5,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="identifier_token",
            full_name="spotify.login5.v3.LoginResponse.identifier_token",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="user_info",
            full_name="spotify.login5.v3.LoginResponse.user_info",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[
        _LOGINRESPONSE_WARNINGS,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1538,
    serialized_end=1914,
)

_CHALLENGES.fields_by_name["challenges"].message_type = _CHALLENGE
_CHALLENGE.fields_by_name["hashcash"].message_type = (
    spotify_dot_login5_dot_v3_dot_challenges_dot_hashcash__pb2.
    _HASHCASHCHALLENGE)
_CHALLENGE.fields_by_name[
    "code"].message_type = spotify_dot_login5_dot_v3_dot_challenges_dot_code__pb2._CODECHALLENGE
_CHALLENGESOLUTIONS.fields_by_name[
    "solutions"].message_type = _CHALLENGESOLUTION
_CHALLENGESOLUTION.fields_by_name["hashcash"].message_type = (
    spotify_dot_login5_dot_v3_dot_challenges_dot_hashcash__pb2.
    _HASHCASHSOLUTION)
_CHALLENGESOLUTION.fields_by_name[
    "code"].message_type = spotify_dot_login5_dot_v3_dot_challenges_dot_code__pb2._CODESOLUTION
_LOGINREQUEST.fields_by_name[
    "client_info"].message_type = spotify_dot_login5_dot_v3_dot_client__info__pb2._CLIENTINFO
_LOGINREQUEST.fields_by_name[
    "challenge_solutions"].message_type = _CHALLENGESOLUTIONS
_LOGINREQUEST.fields_by_name["stored_credential"].message_type = (
    spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2.
    _STOREDCREDENTIAL)
_LOGINREQUEST.fields_by_name["password"].message_type = (
    spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2._PASSWORD)
_LOGINREQUEST.fields_by_name["facebook_access_token"].message_type = (
    spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2.
    _FACEBOOKACCESSTOKEN)
_LOGINREQUEST.fields_by_name["phone_number"].message_type = (
    spotify_dot_login5_dot_v3_dot_identifiers_dot_identifiers__pb2._PHONENUMBER
)
_LOGINREQUEST.fields_by_name["one_time_token"].message_type = (
    spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2.
    _ONETIMETOKEN)
_LOGINREQUEST.fields_by_name["parent_child_credential"].message_type = (
    spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2.
    _PARENTCHILDCREDENTIAL)
_LOGINREQUEST.fields_by_name["apple_sign_in_credential"].message_type = (
    spotify_dot_login5_dot_v3_dot_credentials_dot_credentials__pb2.
    _APPLESIGNINCREDENTIAL)
_LOGINRESPONSE.fields_by_name["ok"].message_type = _LOGINOK
_LOGINRESPONSE.fields_by_name["error"].enum_type = _LOGINERROR
_LOGINRESPONSE.fields_by_name["challenges"].message_type = _CHALLENGES
_LOGINRESPONSE.fields_by_name["warnings"].enum_type = _LOGINRESPONSE_WARNINGS
_LOGINRESPONSE.fields_by_name[
    "user_info"].message_type = spotify_dot_login5_dot_v3_dot_user__info__pb2._USERINFO
_LOGINRESPONSE_WARNINGS.containing_type = _LOGINRESPONSE
DESCRIPTOR.message_types_by_name["Challenges"] = _CHALLENGES
DESCRIPTOR.message_types_by_name["Challenge"] = _CHALLENGE
DESCRIPTOR.message_types_by_name["ChallengeSolutions"] = _CHALLENGESOLUTIONS
DESCRIPTOR.message_types_by_name["ChallengeSolution"] = _CHALLENGESOLUTION
DESCRIPTOR.message_types_by_name["LoginRequest"] = _LOGINREQUEST
DESCRIPTOR.message_types_by_name["LoginOk"] = _LOGINOK
DESCRIPTOR.message_types_by_name["LoginResponse"] = _LOGINRESPONSE
DESCRIPTOR.enum_types_by_name["LoginError"] = _LOGINERROR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Challenges = _reflection.GeneratedProtocolMessageType(
    "Challenges",
    (_message.Message, ),
    {
        "DESCRIPTOR": _CHALLENGES,
        "__module__": "spotify.login5.v3.login5_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.Challenges)
    },
)
_sym_db.RegisterMessage(Challenges)

Challenge = _reflection.GeneratedProtocolMessageType(
    "Challenge",
    (_message.Message, ),
    {
        "DESCRIPTOR": _CHALLENGE,
        "__module__": "spotify.login5.v3.login5_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.Challenge)
    },
)
_sym_db.RegisterMessage(Challenge)

ChallengeSolutions = _reflection.GeneratedProtocolMessageType(
    "ChallengeSolutions",
    (_message.Message, ),
    {
        "DESCRIPTOR": _CHALLENGESOLUTIONS,
        "__module__": "spotify.login5.v3.login5_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.ChallengeSolutions)
    },
)
_sym_db.RegisterMessage(ChallengeSolutions)

ChallengeSolution = _reflection.GeneratedProtocolMessageType(
    "ChallengeSolution",
    (_message.Message, ),
    {
        "DESCRIPTOR": _CHALLENGESOLUTION,
        "__module__": "spotify.login5.v3.login5_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.ChallengeSolution)
    },
)
_sym_db.RegisterMessage(ChallengeSolution)

LoginRequest = _reflection.GeneratedProtocolMessageType(
    "LoginRequest",
    (_message.Message, ),
    {
        "DESCRIPTOR": _LOGINREQUEST,
        "__module__": "spotify.login5.v3.login5_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.LoginRequest)
    },
)
_sym_db.RegisterMessage(LoginRequest)

LoginOk = _reflection.GeneratedProtocolMessageType(
    "LoginOk",
    (_message.Message, ),
    {
        "DESCRIPTOR": _LOGINOK,
        "__module__": "spotify.login5.v3.login5_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.LoginOk)
    },
)
_sym_db.RegisterMessage(LoginOk)

LoginResponse = _reflection.GeneratedProtocolMessageType(
    "LoginResponse",
    (_message.Message, ),
    {
        "DESCRIPTOR": _LOGINRESPONSE,
        "__module__": "spotify.login5.v3.login5_pb2"
        # @@protoc_insertion_point(class_scope:spotify.login5.v3.LoginResponse)
    },
)
_sym_db.RegisterMessage(LoginResponse)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: client_token.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from librespot.proto import Connectivity_pb2 as connectivity__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x63lient_token.proto\x12\x1bspotify.clienttoken.http.v0\x1a\x12\x63onnectivity.proto\"\x84\x02\n\x12\x43lientTokenRequest\x12I\n\x0crequest_type\x18\x01 \x01(\x0e\x32\x33.spotify.clienttoken.http.v0.ClientTokenRequestType\x12\x45\n\x0b\x63lient_data\x18\x02 \x01(\x0b\x32..spotify.clienttoken.http.v0.ClientDataRequestH\x00\x12Q\n\x11\x63hallenge_answers\x18\x03 \x01(\x0b\x32\x34.spotify.clienttoken.http.v0.ChallengeAnswersRequestH\x00\x42\t\n\x07request\"\x99\x01\n\x11\x43lientDataRequest\x12\x16\n\x0e\x63lient_version\x18\x01 \x01(\t\x12\x11\n\tclient_id\x18\x02 \x01(\t\x12Q\n\x15\x63onnectivity_sdk_data\x18\x03 \x01(\x0b\x32\x30.spotify.clienttoken.data.v0.ConnectivitySdkDataH\x00\x42\x06\n\x04\x64\x61ta\"g\n\x17\x43hallengeAnswersRequest\x12\r\n\x05state\x18\x01 \x01(\t\x12=\n\x07\x61nswers\x18\x02 \x03(\x0b\x32,.spotify.clienttoken.http.v0.ChallengeAnswer\"\x81\x02\n\x13\x43lientTokenResponse\x12K\n\rresponse_type\x18\x01 \x01(\x0e\x32\x34.spotify.clienttoken.http.v0.ClientTokenResponseType\x12J\n\rgranted_token\x18\x02 \x01(\x0b\x32\x31.spotify.clienttoken.http.v0.GrantedTokenResponseH\x00\x12\x45\n\nchallenges\x18\x03 \x01(\x0b\x32/.spotify.clienttoken.http.v0.ChallengesResponseH\x00\x42\n\n\x08response\"\x1d\n\x0bTokenDomain\x12\x0e\n\x06\x64omain\x18\x01 \x01(\t\"\x9e\x01\n\x14GrantedTokenResponse\x12\r\n\x05token\x18\x01 \x01(\t\x12\x1d\n\x15\x65xpires_after_seconds\x18\x02 \x01(\x05\x12\x1d\n\x15refresh_after_seconds\x18\x03 \x01(\x05\x12\x39\n\x07\x64omains\x18\x04 \x03(\x0b\x32(.spotify.clienttoken.http.v0.TokenDomain\"_\n\x12\x43hallengesResponse\x12\r\n\x05state\x18\x01 \x01(\t\x12:\n\nchallenges\x18\x02 \x03(\x0b\x32&.spotify.clienttoken.http.v0.Challenge\"&\n\x16\x43lientSecretParameters\x12\x0c\n\x04salt\x18\x01 \x01(\t\"7\n\x14\x45valuateJSParameters\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x11\n\tlibraries\x18\x02 \x03(\t\"4\n\x12HashCashParameters\x12\x0e\n\x06length\x18\x01 \x01(\x05\x12\x0e\n\x06prefix\x18\x02 \x01(\t\"\xda\x02\n\tChallenge\x12\x38\n\x04type\x18\x01 \x01(\x0e\x32*.spotify.clienttoken.http.v0.ChallengeType\x12W\n\x18\x63lient_secret_parameters\x18\x02 \x01(\x0b\x32\x33.spotify.clienttoken.http.v0.ClientSecretParametersH\x00\x12S\n\x16\x65valuate_js_parameters\x18\x03 \x01(\x0b\x32\x31.spotify.clienttoken.http.v0.EvaluateJSParametersH\x00\x12W\n\x1c\x65valuate_hashcash_parameters\x18\x04 \x01(\x0b\x32/.spotify.clienttoken.http.v0.HashCashParametersH\x00\x42\x0c\n\nparameters\"&\n\x16\x43lientSecretHMACAnswer\x12\x0c\n\x04hmac\x18\x01 \x01(\t\"\"\n\x10\x45valuateJSAnswer\x12\x0e\n\x06result\x18\x01 \x01(\t\" \n\x0eHashCashAnswer\x12\x0e\n\x06suffix\x18\x01 \x01(\t\"\xb4\x02\n\x0f\x43hallengeAnswer\x12\x41\n\rChallengeType\x18\x01 \x01(\x0e\x32*.spotify.clienttoken.http.v0.ChallengeType\x12L\n\rclient_secret\x18\x02 \x01(\x0b\x32\x33.spotify.clienttoken.http.v0.ClientSecretHMACAnswerH\x00\x12\x44\n\x0b\x65valuate_js\x18\x03 \x01(\x0b\x32-.spotify.clienttoken.http.v0.EvaluateJSAnswerH\x00\x12@\n\thash_cash\x18\x04 \x01(\x0b\x32+.spotify.clienttoken.http.v0.HashCashAnswerH\x00\x42\x08\n\x06\x61nswer\"(\n\x15\x43lientTokenBadRequest\x12\x0f\n\x07message\x18\x01 \x01(\t*u\n\x16\x43lientTokenRequestType\x12\x13\n\x0fREQUEST_UNKNOWN\x10\x00\x12\x1f\n\x1bREQUEST_CLIENT_DATA_REQUEST\x10\x01\x12%\n!REQUEST_CHALLENGE_ANSWERS_REQUEST\x10\x02*v\n\x17\x43lientTokenResponseType\x12\x14\n\x10RESPONSE_UNKNOWN\x10\x00\x12#\n\x1fRESPONSE_GRANTED_TOKEN_RESPONSE\x10\x01\x12 \n\x1cRESPONSE_CHALLENGES_RESPONSE\x10\x02*|\n\rChallengeType\x12\x15\n\x11\x43HALLENGE_UNKNOWN\x10\x00\x12 \n\x1c\x43HALLENGE_CLIENT_SECRET_HMAC\x10\x01\x12\x19\n\x15\x43HALLENGE_EVALUATE_JS\x10\x02\x12\x17\n\x13\x43HALLENGE_HASH_CASH\x10\x03\x42#\n\x1f\x63om.spotify.clienttoken.http.v0H\x02\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'client_token_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\037com.spotify.clienttoken.http.v0H\002'
  _CLIENTTOKENREQUESTTYPE._serialized_start=2107
  _CLIENTTOKENREQUESTTYPE._serialized_end=2224
  _CLIENTTOKENRESPONSETYPE._serialized_start=2226
  _CLIENTTOKENRESPONSETYPE._serialized_end=2344
  _CHALLENGETYPE._serialized_start=2346
  _CHALLENGETYPE._serialized_end=2470
  _CLIENTTOKENREQUEST._serialized_start=72
  _CLIENTTOKENREQUEST._serialized_end=332
  _CLIENTDATAREQUEST._serialized_start=335
  _CLIENTDATAREQUEST._serialized_end=488
  _CHALLENGEANSWERSREQUEST._serialized_start=490
  _CHALLENGEANSWERSREQUEST._serialized_end=593
  _CLIENTTOKENRESPONSE._serialized_start=596
  _CLIENTTOKENRESPONSE._serialized_end=853
  _TOKENDOMAIN._serialized_start=855
  _TOKENDOMAIN._serialized_end=884
  _GRANTEDTOKENRESPONSE._serialized_start=887
  _GRANTEDTOKENRESPONSE._serialized_end=1045
  _CHALLENGESRESPONSE._serialized_start=1047
  _CHALLENGESRESPONSE._serialized_end=1142
  _CLIENTSECRETPARAMETERS._serialized_start=1144
  _CLIENTSECRETPARAMETERS._serialized_end=1182
  _EVALUATEJSPARAMETERS._serialized_start=1184
  _EVALUATEJSPARAMETERS._serialized_end=1239
  _HASHCASHPARAMETERS._serialized_start=1241
  _HASHCASHPARAMETERS._serialized_end=1293
  _CHALLENGE._serialized_start=1296
  _CHALLENGE._serialized_end=1642
  _CLIENTSECRETHMACANSWER._serialized_start=1644
  _CLIENTSECRETHMACANSWER._serialized_end=1682
  _EVALUATEJSANSWER._serialized_start=1684
  _EVALUATEJSANSWER._serialized_end=1718
  _HASHCASHANSWER._serialized_start=1720
  _HASHCASHANSWER._serialized_end=1752
  _CHALLENGEANSWER._serialized_start=1755
  _CHALLENGEANSWER._serialized_end=2063
  _CLIENTTOKENBADREQUEST._serialized_start=2065
  _CLIENTTOKENBADREQUEST._serialized_end=2105
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
"""
Test script using Zeroconf to verify Login5 authentication
This method doesn't require Spotify Premium
"""

import logging
import time
import pathlib
from librespot.zeroconf import ZeroconfServer

# Enable debug logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

def test_zeroconf_login5():
    """Test Login5 using Zeroconf authentication"""
    print("=== Testing Login5 with Zeroconf ===")
    print("1. Open Spotify on your phone/computer")
    print("2. Look for 'librespot-spotizerr' in Spotify Connect devices")
    print("3. Transfer playback to it")
    print("4. Wait for credentials to be stored...")
    print("\nWaiting for Spotify Connect transfer...")
    
    zs = ZeroconfServer.Builder().create()
    
    start_time = time.time()
    timeout = 60  # 60 seconds timeout
    
    while True:
        time.sleep(1)
        elapsed = time.time() - start_time
        
        if elapsed > timeout:
            print(f"\n⚠ Timeout after {timeout} seconds")
            print("Make sure you:")
            print("- Have Spotify open on another device")
            print("- Can see 'librespot-spotizerr' in Spotify Connect")
            print("- Transfer playback to it")
            return False
            
        if zs._ZeroconfServer__session:
            session = zs._ZeroconfServer__session
            print(f"\n✓ Got session for user: {session.username()}")
            
            # Test token retrieval
            try:
                token_provider = session.tokens()
                token = token_provider.get("playlist-read")
                print(f"✓ Got playlist-read token: {token[:20]}...")
                
                # Test Login5 token
                login5_token = session.get_login5_token()
                if login5_token:
                    print(f"✓ Login5 token available: {login5_token[:20]}...")
                else:
                    print("⚠ Login5 token not available, using fallback")
                
                # Check if credentials were saved
                if pathlib.Path("credentials.json").exists():
                    print("✓ Credentials saved to credentials.json")
                    print("\nYou can now use the stored credentials for future tests!")
                    return True
                else:
                    print("⚠ Credentials not saved")
                    return True
                    
            except Exception as e:
                print(f"✗ Token test failed: {e}")
                return False

def main():
    print("Zeroconf Login5 Test")
    print("=" * 30)
    
    success = test_zeroconf_login5()
    
    print("\n" + "=" * 30)
    if success:
        print("🎉 Zeroconf Login5 test completed successfully!")
        return 0
    else:
        print("⚠ Zeroconf test failed or timed out")
        return 1

if __name__ == "__main__":
    import sys
    sys.exit(main())


from __future__ import annotations
from librespot import util
from librespot.proto.ContextTrack_pb2 import ContextTrack
from librespot.util import Base62
import re


class SpotifyId:
    STATIC_FROM_URI = "fromUri"
    STATIC_FROM_BASE62 = "fromBase62"
    STATIC_FROM_HEX = "fromHex"

    @staticmethod
    def from_base62(base62: str):
        raise NotImplementedError

    @staticmethod
    def from_hex(hex_str: str):
        raise NotImplementedError

    @staticmethod
    def from_uri(uri: str):
        raise NotImplementedError

    def to_spotify_uri(self) -> str:
        raise NotImplementedError

    class SpotifyIdParsingException(Exception):
        pass


class PlayableId:
    base62 = Base62.create_instance_with_inverted_character_set()

    @staticmethod
    def from_uri(uri: str) -> PlayableId:
        if not PlayableId.is_supported(uri):
            return UnsupportedId(uri)
        if TrackId.pattern.search(uri) is not None:
            return TrackId.from_uri(uri)
        if EpisodeId.pattern.search(uri) is not None:
            return EpisodeId.from_uri(uri)
        raise TypeError("Unknown uri: {}".format(uri))

    @staticmethod
    def is_supported(uri: str):
        return (not uri.startswith("spotify:local:")
                and not uri == "spotify:delimiter"
                and not uri == "spotify:meta:delimiter")

    @staticmethod
    def should_play(track: ContextTrack):
        return track.metadata_or_default

    def get_gid(self) -> bytes:
        raise NotImplementedError

    def hex_id(self) -> str:
        raise NotImplementedError

    def to_spotify_uri(self) -> str:
        raise NotImplementedError


class PlaylistId(SpotifyId):
    base62 = Base62.create_instance_with_inverted_character_set()
    pattern = re.compile(r"spotify:playlist:(.{22})")
    __id: str

    def __init__(self, _id: str):
        self.__id = _id

    @staticmethod
    def from_uri(uri: str) -> PlaylistId:
        matcher = PlaylistId.pattern.search(uri)
        if matcher is not None:
            playlist_id = matcher.group(1)
            return PlaylistId(playlist_id)
        raise TypeError("Not a Spotify playlist ID: {}.".format(uri))

    def id(self) -> str:
        return self.__id

    def to_spotify_uri(self) -> str:
        return "spotify:playlist:" + self.__id


class UnsupportedId(PlayableId):
    uri: str

    def __init__(self, uri: str):
        self.uri = uri

    def get_gid(self) -> bytes:
        raise TypeError()

    def hex_id(self) -> str:
        raise TypeError()

    def to_spotify_uri(self) -> str:
        return self.uri


class AlbumId(SpotifyId):
    base62 = Base62.create_instance_with_inverted_character_set()
    pattern = re.compile(r"spotify:album:(.{22})")
    __hex_id: str

    def __init__(self, hex_id: str):
        self.__hex_id = hex_id.lower()

    @staticmethod
    def from_uri(uri: str) -> AlbumId:
        matcher = AlbumId.pattern.search(uri)
        if matcher is not None:
            album_id = matcher.group(1)
            return AlbumId(util.bytes_to_hex(AlbumId.base62.decode(album_id.encode(), 16)))
        raise TypeError("Not a Spotify album ID: {}.".format(uri))

    @staticmethod
    def from_base62(base62: str) -> AlbumId:
        return AlbumId(util.bytes_to_hex(AlbumId.base62.decode(base62.encode(), 16)))

    @staticmethod
    def from_hex(hex_str: str) -> AlbumId:
        return AlbumId(hex_str)

    def to_mercury_uri(self) -> str:
        return "hm://metadata/4/album/{}".format(self.__hex_id)

    def hex_id(self) -> str:
        return self.__hex_id

    def to_spotify_uri(self) -> str:
        return "spotify:album:{}".format(
            AlbumId.base62.encode(util.hex_to_bytes(self.__hex_id)).decode())


class ArtistId(SpotifyId):
    base62 = Base62.create_instance_with_inverted_character_set()
    pattern = re.compile("spotify:artist:(.{22})")
    __hex_id: str

    def __init__(self, hex_id: str):
        self.__hex_id = hex_id.lower()

    @staticmethod
    def from_uri(uri: str) -> ArtistId:
        matcher = ArtistId.pattern.search(uri)
        if matcher is not None:
            artist_id = matcher.group(1)
            return ArtistId(
                util.bytes_to_hex(ArtistId.base62.decode(artist_id.encode(), 16)))
        raise TypeError("Not a Spotify artist ID: {}".format(uri))

    @staticmethod
    def from_base62(base62: str) -> ArtistId:
        return ArtistId(util.bytes_to_hex(ArtistId.base62.decode(base62.encode(), 16)))

    @staticmethod
    def from_hex(hex_str: str) -> ArtistId:
        return ArtistId(hex_str)

    def to_mercury_uri(self) -> str:
        return "hm://metadata/4/artist/{}".format(self.__hex_id)

    def to_spotify_uri(self) -> str:
        return "spotify:artist:{}".format(
            ArtistId.base62.encode(util.hex_to_bytes(self.__hex_id)).decode())

    def hex_id(self) -> str:
        return self.__hex_id


class EpisodeId(SpotifyId, PlayableId):
    pattern = re.compile(r"spotify:episode:(.{22})")
    __hex_id: str

    def __init__(self, hex_id: str):
        self.__hex_id = hex_id.lower()

    @staticmethod
    def from_uri(uri: str) -> EpisodeId:
        matcher = EpisodeId.pattern.search(uri)
        if matcher is not None:
            episode_id = matcher.group(1)
            return EpisodeId(
                util.bytes_to_hex(PlayableId.base62.decode(episode_id.encode(), 16)))
        raise TypeError("Not a Spotify episode ID: {}".format(uri))

    @staticmethod
    def from_base62(base62: str) -> EpisodeId:
        return EpisodeId(
            util.bytes_to_hex(PlayableId.base62.decode(base62.encode(), 16)))

    @staticmethod
    def from_hex(hex_str: str) -> EpisodeId:
        return EpisodeId(hex_str)

    def to_mercury_uri(self) -> str:
        return "hm://metadata/4/episode/{}".format(self.__hex_id)

    def to_spotify_uri(self) -> str:
        return "Spotify:episode:{}".format(
            PlayableId.base62.encode(util.hex_to_bytes(self.__hex_id)).decode())

    def hex_id(self) -> str:
        return self.__hex_id

    def get_gid(self) -> bytes:
        return util.hex_to_bytes(self.__hex_id)


class ShowId(SpotifyId):
    base62 = Base62.create_instance_with_inverted_character_set()
    pattern = re.compile("spotify:show:(.{22})")
    __hex_id: str

    def __init__(self, hex_id: str):
        self.__hex_id = hex_id

    @staticmethod
    def from_uri(uri: str) -> ShowId:
        matcher = ShowId.pattern.search(uri)
        if matcher is not None:
            show_id = matcher.group(1)
            return ShowId(util.bytes_to_hex(ShowId.base62.decode(show_id.encode(), 16)))
        raise TypeError("Not a Spotify show ID: {}".format(uri))

    @staticmethod
    def from_base62(base62: str) -> ShowId:
        return ShowId(util.bytes_to_hex(ShowId.base62.decode(base62.encode(), 16)))

    @staticmethod
    def from_hex(hex_str: str) -> ShowId:
        return ShowId(hex_str)

    def to_mercury_uri(self) -> str:
        return "hm://metadata/4/show/{}".format(self.__hex_id)

    def to_spotify_uri(self) -> str:
        return "spotify:show:{}".format(
            ShowId.base62.encode(util.hex_to_bytes(self.__hex_id)).decode())

    def hex_id(self) -> str:
        return self.__hex_id


class TrackId(PlayableId, SpotifyId):
    pattern = re.compile("spotify:track:(.{22})")
    __hex_id: str

    def __init__(self, hex_id: str):
        self.__hex_id = hex_id.lower()

    @staticmethod
    def from_uri(uri: str) -> TrackId:
        search = TrackId.pattern.search(uri)
        if search is not None:
            track_id = search.group(1)
            return TrackId(
                util.bytes_to_hex(PlayableId.base62.decode(track_id.encode(), 16)))
        raise RuntimeError("Not a Spotify track ID: {}".format(uri))

    @staticmethod
    def from_base62(base62: str) -> TrackId:
        return TrackId(util.bytes_to_hex(PlayableId.base62.decode(base62.encode(), 16)))

    @staticmethod
    def from_hex(hex_str: str) -> TrackId:
        return TrackId(hex_str)

    def to_mercury_uri(self) -> str:
        return "hm://metadata/4/track/{}".format(self.__hex_id)

    def to_spotify_uri(self) -> str:
        return "spotify:track:{}".format(TrackId.base62.encode(util.hex_to_bytes(self.__hex_id)).decode())

    def hex_id(self) -> str:
        return self.__hex_id

    def get_gid(self) -> bytes:
        return util.hex_to_bytes(self.__hex_id)
